/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (c) 2024 chudihuang */
#include <argp.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <time.h>
#include <sys/stat.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <err.h>
#include <sys/time.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>
#include "dslower.skel.h"
#include "trace_helpers.h"
#include "dslower.h"
struct syms_cache *syms_cache = NULL;
static volatile sig_atomic_t exiting = 0;
FILE *filep = NULL;
char filename[256] = {0};
struct env
{
        __u64 min_us;
        int pid;
        char comm[TASK_COMM_LEN];
        int comm_len;
        int stack_storage_size;
        int perf_max_stack_depth;
        bool logfile;
        bool verbose;
} env = {
        .pid = 0,
        .comm_len = 0,
        .min_us = 10000,
        .logfile = false,
        .stack_storage_size = 2048,
        .perf_max_stack_depth = 127,
};
int ifd, u_sfd, k_sfd;
unsigned int zero_key = 0;
int stacktrace_map_fd, min_ns_map_fd;
unsigned long long min_ns;
struct ksyms *ksyms = NULL;
const char *argp_program_version = "dslower 0.1";
const char *argp_program_bug_address =
    "https://github.com/iovisor/bcc/tree/master/libbpf-tools";
const char argp_program_doc[] =
    "Trace process block.\n"
    "\n"
    "USAGE: dslower [min_ms] [-p xxx] [-c xxx] [-f xxx.log]\n"
    "\n"
    "EXAMPLES:\n"
    "    dslower               # run forever, detect delay more than 10000us(default)\n"
    "    dslower 1000          # detect delay more than 1000 us, dump stack\n"
    "    dslower -p 185        # trace PID 185(include threads)\n"
    "    noschedule -c test    # trace process who's name is test\n"
    "    dslower -f xxx.log    # log to xxx.log (default output to terminal)\n";

static const struct argp_option opts[] = {
    { "pid", 'p', "PID", 0, "Process PID to trace(include threads)" },
    {"comm", 'c', "COMM", 0, "Name of the application"},
    //  {"stacktrace", 's', NULL, 0, "Show stack trace"},
    {"verbose", 'v', NULL, 0, "Verbose debug output"},
    {"logfile", 'f', "LOGFILE", 0, "logfile for result"},
    {NULL, 'h', NULL, OPTION_HIDDEN, "Show the full help"},
    {},
};

FILE *open_log_file(char *filename)
{
        FILE *filep;

        filep = fopen(filename, "w+");
        if (!filep)
        {
                fprintf(stderr, "%s :fopen %s\n", strerror(errno), filename);
                return NULL;
        }

        return filep;
}

static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
        static int pos_args, pid, len;
        long long min_us;
        switch (key)
        {
        case 'h':
                argp_state_help(state, stderr, ARGP_HELP_STD_HELP);
                break;
        case 'v':
                env.verbose = true;
                break;
        case 'p':
                errno = 0;
                pid = strtol(arg, NULL, 10);
                if (errno || pid <= 0) {
                        warn("Invalid PID: %s\n", arg);
                        // argp_usageڴӡ÷Ϣ˳
                        argp_usage(state);
                }
                env.pid = pid;
                break;
        case 'c':
                // env.comm = arg;
                len = strlen(arg) + 1;
                env.comm_len = len > TASK_COMM_LEN ? TASK_COMM_LEN : len;
                strncpy(env.comm, arg, env.comm_len);
                break;
        case 'f':
                if (strlen(arg) > 255)
                {
                        fprintf(stderr, "filename too long\n");
                        argp_usage(state);
                }
                strncpy(filename, arg, sizeof(filename)-1);
                env.logfile = true;
                break;
        case ARGP_KEY_ARG:
                if (pos_args++)
                {
                        fprintf(stderr,
                                "Unrecognized positional argument: %s\n", arg);
                        argp_usage(state);
                }
                errno = 0;
                min_us = strtoll(arg, NULL, 10);
                if (errno || min_us <= 0)
                {
                        fprintf(stderr, "Invalid delay (in us): %s\n", arg);
                        argp_usage(state);
                }
                env.min_us = min_us;
                break;
        default:
                return ARGP_ERR_UNKNOWN;
        }
        return 0;
}

static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
{
        if (level == LIBBPF_DEBUG && !env.verbose)
                return 0;
        return vfprintf(stderr, format, args);
}

static void sig_int(int signo)
{
        exiting = 1;
}
void print_user_stack(FILE *filep, int perf_max_stack_depth, int sfd, bool logfile, int u_stack_id, unsigned long pid, struct syms_cache *syms_cache)
{
        const struct syms *syms;
        const struct sym *sym;
        int i;
        unsigned long *ip;

        ip = calloc(perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        fprintf(stderr, "failed to alloc ip\n");
                return;
        }

        if (bpf_map_lookup_elem(sfd, &u_stack_id, ip) != 0)
        {
                if (logfile)
                        fprintf(filep, "    [Missed User Stack:%d]\n", u_stack_id);
                else
                        fprintf(stderr, "    [Missed User Stack:%d]\n", u_stack_id);
                goto skip_user_stack;
        }

        syms = syms_cache__get_syms(syms_cache, pid);
        if (!syms)
        {
                if (logfile)
                        fprintf(filep, "    failed to get syms\n");
                else
                        fprintf(stderr, "    failed to get syms\n");
                goto skip_user_stack;
        }
	for (i = 0; i < perf_max_stack_depth && ip[i]; i++)
        {
                sym = syms__map_addr(syms, ip[i]);
                if (sym)
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx %s\n", ip[i], sym->name);
                        else
                                printf("    0x%lx %s\n", ip[i], sym->name);
                }
                else
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx [unknown]\n", ip[i]);
                        else
                                printf("    0x%lx [unknown]\n", ip[i]);
                }
        }

skip_user_stack:
        free(ip);
}
static void print_blocker_stack(const struct event *e, struct ksyms *ksyms)
{
        const struct ksym *ksym;
        int i;
        unsigned long *ip;
        ip = calloc(env.perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (env.logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        printf("failed to alloc ip\n");
                return;
        }
        if (env.logfile)
                fprintf(filep, "blocker stack:\n");
        else
                printf("blocker stack:\n");
        if (bpf_map_lookup_elem(k_sfd, &e->blocker_kstack_id, ip) != 0)
        {
                if (env.logfile)
                        fprintf(filep, "    [Missed blocker Kernel StackID:%d]\n", e->blocker_kstack_id);
                else
                        printf("    [Missed blocker Kernel StackID:%d]\n", e->blocker_kstack_id);
                goto skip_stack;
        }
        for (i = 0; i < env.perf_max_stack_depth && ip[i]; i++)
        {
                ksym = ksyms__map_addr(ksyms, ip[i]);
                if (ksym)
                {
                        if (env.logfile)
                                fprintf(filep, "    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                        else
                                printf("    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                }
                else
                {
                        if (env.logfile)
                                fprintf(filep, "    %s\n", "Unknown");
                        else
                                printf("    %s\n", "Unknown");
                }
        }
skip_stack:
        free(ip);

        print_user_stack(filep, env.perf_max_stack_depth, u_sfd, env.logfile, e->blocker_ustack_id, e->blocker_tgid, syms_cache);
}

static void print_waker_stack(const struct event *e, struct ksyms *ksyms)
{
        const struct ksym *ksym;
        int i;
        unsigned long *ip;
        ip = calloc(env.perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (env.logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        printf("failed to alloc ip\n");
                return;
        }
        if (env.logfile)
                fprintf(filep, "waker stack:\n");
        else
                printf("waker stack:\n");
        if (bpf_map_lookup_elem(k_sfd, &e->waker_kstack_id, ip) != 0)
        {
                if (env.logfile)
                        fprintf(filep, "    [Missed waker Kernel StackID:%d]\n", e->waker_kstack_id);
                else
                        printf("    [Missed waker Kernel StackID:%d]\n", e->waker_kstack_id);
                goto skip_stack;
        }
        for (i = 0; i < env.perf_max_stack_depth && ip[i]; i++)
        {
                ksym = ksyms__map_addr(ksyms, ip[i]);
                if (ksym)
                {
                        if (env.logfile)
                                fprintf(filep, "    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                        else
                                printf("    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                }
                else
                {
                        if (env.logfile)
                                fprintf(filep, "    %s\n", "Unknown");
                        else
                                printf("    %s\n", "Unknown");
                }
        }
skip_stack:
        free(ip);

        print_user_stack(filep, env.perf_max_stack_depth, u_sfd, env.logfile, e->waker_ustack_id, e->waker_tgid, syms_cache);
}
void handle_event(void *ctx, int cpu, void *data, __u32 data_sz)
{
        const struct event *e = data;
        struct tm *tm;
        char ts[64]; // Increase the size of the ts array
        struct timeval tv;
        time_t t;

        gettimeofday(&tv, NULL);
        t = tv.tv_sec;
        tm = localtime(&t);
        strftime(ts, sizeof(ts) - 6, "%F_%H:%M:%S", tm);       // Leave 6 bytes for milliseconds and null terminator
        sprintf(ts + strlen(ts), ".%03ld", tv.tv_usec / 1000); // Append milliseconds using sprintf
        if (env.logfile)
                fprintf(filep, "[%s]: \n[waker]  : pid:%-8d comm: %-16s runtime:%-10lld(us) runcount:%-lld   \n[blocker]: pid:%-8d comm: %-16s blocked:%-10lld(us) runcount:%-lld\n",
                        ts, e->waker_pid, e->waker_task, (e->waker_runtime / 1000), (e->waker_run_count), e->blocker_pid, e->blocker_task, (e->delta_ns / 1000), (e->blocker_run_count));
        else
                printf("[%s]: \n[waker]  : pid:%-8d comm: %-16s runtime:%-10lld(us) runcount:%-lld   \n[blocker]: pid:%-8d comm: %-16s blocked:%-10lld(us) runcount:%-lld\n",
                       ts, e->waker_pid, e->waker_task, (e->waker_runtime / 1000), (e->waker_run_count), e->blocker_pid, e->blocker_task, (e->delta_ns / 1000), (e->blocker_run_count));
        print_waker_stack(e, ksyms);
        print_blocker_stack(e, ksyms);
        if (env.logfile)
                fflush(filep);
}

void handle_lost_events(void *ctx, int cpu, __u64 lost_cnt)
{
        printf("Lost %llu events on CPU #%d!\n", lost_cnt, cpu);
}

int main(int argc, char **argv)
{
        static const struct argp argp = {
            .options = opts,
            .parser = parse_arg,
            .doc = argp_program_doc,
        };
        struct perf_buffer *pb = NULL;
        struct dslower_bpf *obj;
        int err, arg_fd;
        struct arg_val arg_val = {};

        setlinebuf(stdout);
        memset(env.comm, '\0', TASK_COMM_LEN);
        err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
        if (err)
                return err;

        if(env.logfile)
        {
                filep = open_log_file(filename);
                if (!filep)
                {
                        fprintf(stderr, "Failed to open log file\n");
                        return 0;
                }
        }

        libbpf_set_strict_mode(LIBBPF_STRICT_ALL);
        libbpf_set_print(libbpf_print_fn);

        obj = dslower_bpf__open();
        if (!obj)
        {
                fprintf(stderr, "failed to open BPF object\n");
                return 1;
        }
        bpf_map__set_value_size(obj->maps.user_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
        bpf_map__set_max_entries(obj->maps.user_stackmap, env.stack_storage_size);
        bpf_map__set_value_size(obj->maps.kernel_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
        bpf_map__set_max_entries(obj->maps.kernel_stackmap, env.stack_storage_size);
        err = dslower_bpf__load(obj);
        if (err)
        {
                fprintf(stderr, "failed to load BPF object: %d\n", err);
                goto cleanup;
        }

        ksyms = ksyms__load();
        if (!ksyms)
        {
                fprintf(stderr, "failed to load kallsyms\n");
                goto cleanup;
        }

        syms_cache = syms_cache__new(0);
        if (!syms_cache)
        {
                fprintf(stderr, "failed to create syms_cache\n");
                goto cleanup;
        }

        u_sfd = bpf_map__fd(obj->maps.user_stackmap);
        k_sfd = bpf_map__fd(obj->maps.kernel_stackmap);
        arg_val.threshold = env.min_us * 1000;
        arg_val.tgid = env.pid;
        if (env.comm_len && env.comm)
        {
                strncpy((char *)arg_val.targ_comm, env.comm, env.comm_len);
                arg_val.comm_len = env.comm_len;
        }
        arg_fd = bpf_map__fd(obj->maps.arg_map);
        err = bpf_map_update_elem(arg_fd, &zero_key, &arg_val, 0);
        if (err)
        {
                fprintf(stderr, "Failed to update arg_map\n");
                goto cleanup;
        }
        err = dslower_bpf__attach(obj);
        if (err)
        {
                fprintf(stderr, "failed to attach BPF programs\n");
                goto cleanup;
        }
        if (env.logfile)
        {
                fprintf(filep, "Tracing process blocked than %llu us\n", env.min_us);
                fflush(filep);
        }
        else
        {
                printf("Tracing process blocked than %llu us\n", env.min_us);
        }
        pb = perf_buffer__new(bpf_map__fd(obj->maps.events), 64,
                              handle_event, handle_lost_events, NULL, NULL);
        if (!pb)
        {
                err = -errno;
                fprintf(stderr, "failed to open perf buffer: %d\n", err);
                goto cleanup;
        }

        if (signal(SIGINT, sig_int) == SIG_ERR)
        {
                fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
                err = 1;
                goto cleanup;
        }

        if (signal(SIGTERM, sig_int) == SIG_ERR)
        {
                fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
                err = 1;
                goto cleanup;
        }


        while (!exiting)
        {
                err = perf_buffer__poll(pb, -1);
                if (err < 0 && err != -EINTR)
                {
                        fprintf(stderr, "error polling perf buffer: %s\n", strerror(-err));
                        goto cleanup;
                }
                /* reset err to return 0 if exiting */
                err = 0;
        }

cleanup:
        perf_buffer__free(pb);
        dslower_bpf__destroy(obj);
        if (ksyms)
                ksyms__free(ksyms);
        if (syms_cache)
                syms_cache__free(syms_cache);
        if (env.logfile && filep)
        {
                fflush(filep);
                fclose(filep);
        }
        return err != 0;
}
