#include <argp.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <time.h>
#include <bpf/btf.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>
#include "dtop.skel.h"
#include "trace_helpers.h"
#include "dtop.h"
#define warn(...) fprintf(stderr, __VA_ARGS__)
#define OUTPUT_ROWS_LIMIT 10240
static volatile sig_atomic_t exiting = 0;
static int interval = 1;
static int count = 99999999;
static bool clear_screen = true;
static int output_rows = 20;
static bool verbose = false;
const char *argp_program_version = "dtop 0.1";
const char *argp_program_bug_address =
        "https://github.com/iovisor/bcc/tree/master/libbpf-tools";
const char argp_program_doc[] =
"Trace process blocktime per interval sec.\n"
"\n"
"USAGE: dtop [--help] [-C] [-r] [interval] [count]\n"
"\n"
"EXAMPLES:\n"
"    dtop           # process blocktime top,refresh default every 1s sec\n"
"    dtop -C        # don't clear the screen\n"
"    dtop 5         # 5 second summaries\n"
"    dtop 5 10      # 5 second summaries, 10 times only"
"    dtop -r 100    # 100 rows to print, default 20\n";

static const struct argp_option opts[] = {
	{ "rows", 'r', "ROWS", 0, "maximum rows to print, default 20" },
        { "noclear", 'C', NULL, 0, "Show stack trace"},
        { "verbose", 'v', NULL, 0, "Verbose debug output" },
        { NULL, 'h', NULL, OPTION_HIDDEN, "Show the full help" },
        {},
};

static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
        long rows;
        static int pos_args;

        switch (key) {
        case 'C':
                clear_screen = false;
                break;
        case 'r':
                errno = 0;
                rows = strtol(arg, NULL, 10);
                if (errno || rows <= 0) {
                        warn("invalid rows: %s\n", arg);
                        argp_usage(state);
                }
                output_rows = rows;
                if (output_rows > OUTPUT_ROWS_LIMIT)
                        output_rows = OUTPUT_ROWS_LIMIT;
                break;
        case 'v':
                verbose = true;
                break;
        case 'h':
                argp_state_help(state, stderr, ARGP_HELP_STD_HELP);
                break;
        case ARGP_KEY_ARG:
                errno = 0;
                if (pos_args == 0) {
                        interval = strtol(arg, NULL, 10);
                        if (errno || interval <= 0) {
                                warn("invalid interval\n");
                                argp_usage(state);
                        }
                } else if (pos_args == 1) {
                        count = strtol(arg, NULL, 10);
                        if (errno || count <= 0) {
                                warn("invalid count\n");
                                argp_usage(state);
                        }
                } else {
                        warn("unrecognized positional argument: %s\n", arg);
                        argp_usage(state);
                }
                pos_args++;
                break;
        default:
                return ARGP_ERR_UNKNOWN;
        }
        return 0;
}


static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
{
        if (level == LIBBPF_DEBUG && !verbose)
                return 0;
        return vfprintf(stderr, format, args);
}

static void sig_int(int signo)
{
        exiting = 1;
}


struct data_t {
        struct task_t key;
        struct val_t value;
};

static int sort_column(const void *obj1, const void *obj2)
{
        struct data_t *d1 = (struct data_t *) obj1;
        struct data_t *d2 = (struct data_t *) obj2;

        struct val_t *s1 = &d1->value;
        struct val_t *s2 = &d2->value;
        return s2->ns - s1->ns;
}
static int print_stat(struct dtop_bpf *obj)
{
        FILE *f;
        time_t t;
        struct tm *tm;
        struct timeval tv;
        char ts[16], buf[256];
        struct task_t *prev_key = NULL;
        static struct data_t datas[OUTPUT_ROWS_LIMIT];
        int n, i, err = 0, rows = 0;
        int fd = bpf_map__fd(obj->maps.counts);
        f = fopen("/proc/loadavg", "r");
        if (f) {
                gettimeofday(&tv, NULL);
                t = tv.tv_sec;
                tm = localtime(&t);
                strftime(ts, sizeof(ts), "%H:%M:%S", tm);
                sprintf(ts + strlen(ts), ".%03ld", tv.tv_usec / 1000);
                memset(buf, 0, sizeof(buf));
                n = fread(buf, 1, sizeof(buf), f);
                if (n)
                        printf("%8s loadavg: %s\n", ts, buf);
                fclose(f);
        }
        printf("%-16s  %-7s  %s\n","COMM", "PID", "BLOCK TIME(usecs)");
        while (1) {
                err = bpf_map_get_next_key(fd, prev_key, &datas[rows].key);
                if (err) {
                        if (errno == ENOENT) {
                                err = 0;
                                break;
                        }
                        warn("bpf_map_get_next_key failed: %s\n", strerror(errno));
                        return err;
                }
                err = bpf_map_lookup_elem(fd, &datas[rows].key, &datas[rows].value);
                if (err) {
                        warn("bpf_map_lookup_elem failed: %s\n", strerror(errno));
                        return err;
                }
                prev_key = &datas[rows].key;
                rows++;
        }

        qsort(datas, rows, sizeof(struct data_t), sort_column);
        rows = rows < output_rows ? rows : output_rows;
        for (i = 0; i < rows; i++) {
		struct task_t *key = &datas[i].key;
                struct val_t *value = &datas[i].value;
                float avg_us = 0; 
                avg_us = ((float) value->ns) / 1000;
                printf("%-16s  %-7d  %6.2f\n",key->comm,key->pid,avg_us);

        }

        printf("\n");
        prev_key = NULL;
        while (1) {
                struct task_t key;

                err = bpf_map_get_next_key(fd, prev_key, &key);
                if (err) {
                        if (errno == ENOENT) {
                                err = 0;
                                break;
                        }
                        warn("bpf_map_get_next_key failed: %s\n", strerror(errno));
                        return err;
                }
                err = bpf_map_delete_elem(fd, &key);
                if (err) {
                        warn("bpf_map_delete_elem failed: %s\n", strerror(errno));
                        return err;
                }
                prev_key = &key;
        }
        return err;
}


int main(int argc, char **argv)
{
        static const struct argp argp = {
                .options = opts,
                .parser = parse_arg,
                .doc = argp_program_doc,
        };
        struct dtop_bpf *obj;
        int err;

        setlinebuf(stdout);

        err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
        if (err)
                return err;

        libbpf_set_strict_mode(LIBBPF_STRICT_ALL);
        libbpf_set_print(libbpf_print_fn);

        obj = dtop_bpf__open();
        if (!obj) {
                fprintf(stderr, "failed to open BPF object\n");
                return 1;
        }
        err = dtop_bpf__load(obj);
        if (err) {
                fprintf(stderr, "failed to load BPF object: %d\n", err);
                goto cleanup;
        }
      
        err = dtop_bpf__attach(obj);
        if (err) {
                fprintf(stderr, "failed to attach BPF programs\n");
                goto cleanup;
        }
        if (signal(SIGINT, sig_int) == SIG_ERR) {
                fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
                err = 1;
                goto cleanup;
        }

        printf("Tracing... Output every %d secs. Hit Ctrl-C to end\n",interval);
        while (1) {
                sleep(interval);

                if (clear_screen) {
                        err = system("clear");
                        if (err)
                                goto cleanup;
                }

                err = print_stat(obj);
                if (err)
                        goto cleanup;

                count--;
                if (exiting || !count)
                        goto cleanup;
        }


cleanup:
        dtop_bpf__destroy(obj);
        return err != 0;
}
