/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (c) 2024 chudihuang/hilyzhang */
#define BUILD_WITH_BTF 1
#include <argp.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <time.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <err.h>
#include <sys/time.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>
#include <stdbool.h>
#include <sys/stat.h>
#include "irqslower.h"
#include "irqslower.skel.h"
#include "trace_helpers.h"

struct env {
	__u64 min_us;
    int stack_storage_size;
    int perf_max_stack_depth;
    bool stack;
    bool verbose;
    bool logfile;
    bool enable_softirq;
    bool enable_hardirq;
} env = {
	.min_us = 10000,
	.stack_storage_size = 2048,
	.perf_max_stack_depth = 127,
    .stack = false,
    .logfile = false,
    .enable_softirq = false,
    .enable_hardirq = false,
};

static volatile sig_atomic_t exiting = 0;
FILE *filep = NULL;
char filename[256] = {0};
int u_sfd, k_sfd;
struct ksyms *ksyms = NULL;
struct syms_cache *syms_cache = NULL;

const char *argp_program_version = "irqslower 0.1";

const char argp_program_doc[] =
"Trace high irq latency.\n"
"\n"
"USAGE: irqslower [--help] [-s] [-S] [-H]\n"
"\n"
"EXAMPLES:\n"
"    irqslower                 #trace run in interrupt higher than default 10ms\n"
"    irqslower  -s             #print stack\n"
"    irqslower 1000 -S         #trace run in softirq higher than 1ms\n"
"    irqslower 1000 -H         #trace run in hardirq higher than 1ms\n"
"    irqslower -f xxx.log      #log to xxx.log, default output to terminal\n";

static const struct argp_option opts[] = {
    {"stack", 's', NULL, 0, "record stack"},
	{"logfile", 'f', "LOGFILE", 0, "logfile for result"},
    {"verbose", 'v', NULL, 0, "Verbose debug output"},
    {"softirq", 'S', NULL, 0, "trace softirq only"},
    {"hardirq", 'H', NULL, 0, "trace hardirq only"},
    {NULL, 'h', NULL, OPTION_HIDDEN, "Show the full help"},
	{},
};

enum {
	HI_SOFTIRQ = 0,
	TIMER_SOFTIRQ = 1,
	NET_TX_SOFTIRQ = 2,
	NET_RX_SOFTIRQ = 3,
	BLOCK_SOFTIRQ = 4,
	IRQ_POLL_SOFTIRQ = 5,
	TASKLET_SOFTIRQ = 6,
	SCHED_SOFTIRQ = 7,
	HRTIMER_SOFTIRQ = 8,
	RCU_SOFTIRQ = 9,
	NR_SOFTIRQS = 10,
};

static char *si_vec_names[] = {
	[HI_SOFTIRQ] = "hi",
	[TIMER_SOFTIRQ] = "timer",
	[NET_TX_SOFTIRQ] = "net_tx",
	[NET_RX_SOFTIRQ] = "net_rx",
	[BLOCK_SOFTIRQ] = "block",
	[IRQ_POLL_SOFTIRQ] = "irq_poll",
	[TASKLET_SOFTIRQ] = "tasklet",
	[SCHED_SOFTIRQ] = "sched",
	[HRTIMER_SOFTIRQ] = "hrtimer",
	[RCU_SOFTIRQ] = "rcu",
};

static char *special_vec_names[] = {
        [CALL_FUNCTION_SINGLE_VECTOR] = "IPI SINGLE",
        [CALL_FUNCTION_VECTOR] = "IPI",
        [ERROR_APIC_VECTOR] = "ERROR APIC",
};

static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
    static int pos_args;
    long long min_us;

    switch (key) {
     case 'h':
         argp_state_help(state, stderr, ARGP_HELP_STD_HELP);
         break;
     case 's':
         env.stack = true;
         break;
      case 'S':
         env.enable_softirq = true;
         break;
      case 'H':
         env.enable_hardirq = true;
         break;
      case 'f':
          if (strlen(arg) > 255)
          {
              fprintf(stderr, "filename too long\n");
              argp_usage(state);
          }
          strncpy(filename, arg, sizeof(filename)-1);
          env.logfile = true;
          break;
      case 'v':
          env.verbose = true;
          break;
      case ARGP_KEY_ARG:
          if (pos_args++)
          {
              fprintf(stderr, "Unrecognized positional argument: %s\n", arg);
              argp_usage(state);
          }
          errno = 0;
          min_us = strtoll(arg, NULL, 10);
          if (errno || min_us <= 0)
          {
              fprintf(stderr, "Invalid delay (in us): %s\n", arg);
              argp_usage(state);
          }
          env.min_us = min_us;
          break;
          default:
              return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

FILE *open_log_file(char *filename)
{
        FILE *filep;

        filep = fopen(filename, "w+");
        if (!filep)
        {
                fprintf(stderr, "%s :fopen %s\n", strerror(errno), filename);
                return NULL;
        }

        return filep;
}

static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
{
	if (level == LIBBPF_DEBUG && !env.verbose)
		return 0;
	return vfprintf(stderr, format, args);
}

static void sig_int(int signo)
{
    exiting = 1;
}

void print_kernel_stack(FILE *filep, int perf_max_stack_depth, int sfd, bool logfile, int k_stack_id, struct ksyms *ksyms)
{
        const struct ksym *ksym;
        int i;
        unsigned long *ip;

        ip = calloc(perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        fprintf(stderr, "failed to alloc ip\n");
                return;
        }

        if (bpf_map_lookup_elem(sfd, &k_stack_id, ip) != 0)
        {
                if (logfile)
                        fprintf(filep, "    [Missed Kernel Stack:%d]\n", k_stack_id);
                else
                        fprintf(stderr, "    [Missed Kernel Stack:%d]\n", k_stack_id);
                goto skip_kernel_stack;
        }
        for (i = 0; i < perf_max_stack_depth && ip[i]; i++)
        {
                ksym = ksyms__map_addr(ksyms, ip[i]);
                if (ksym)
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                        else
                                printf("    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                }
                else
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx [unknown]\n", ip[i]);
                        else
                                printf("    0x%lx [unknown]\n", ip[i]);
                }
        }

skip_kernel_stack:
        free(ip);
}


void print_user_stack(FILE *filep, int perf_max_stack_depth, int sfd, bool logfile, int u_stack_id, unsigned long tgid, struct syms_cache *syms_cache)
{
        const struct syms *syms;
        const struct sym *sym;
        int i;
        unsigned long *ip;

        ip = calloc(perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        fprintf(stderr, "failed to alloc ip\n");
                return;
        }

        if (bpf_map_lookup_elem(sfd, &u_stack_id, ip) != 0)
        {
                if (logfile)
                        fprintf(filep, "    [Missed User Stack:%d]\n", u_stack_id);
                else
                        fprintf(stderr, "    [Missed User Stack:%d]\n", u_stack_id);
                goto skip_user_stack;
        }

        syms = syms_cache__get_syms(syms_cache, tgid);
        if (!syms)
        {
                if (logfile)
                        fprintf(filep, "    failed to get syms\n");
                else
                        fprintf(stderr, "    failed to get syms\n");
                goto skip_user_stack;
        }
	for (i = 0; i < perf_max_stack_depth && ip[i]; i++)
        {
                sym = syms__map_addr(syms, ip[i]);
                if (sym)
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx %s\n", ip[i], sym->name);
                        else
                                printf("    0x%lx %s\n", ip[i], sym->name);
                }
                else
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx [unknown]\n", ip[i]);
                        else
                                printf("    0x%lx [unknown]\n", ip[i]);
                }
        }

skip_user_stack:
        free(ip);
}

void handle_event(void *ctx, int cpu, void *data, __u32 data_sz)
{
    const struct event *e = data;
    struct tm *tm;
    char ts[64]; // Increase the size of the ts array
    struct timeval tv;
    time_t t;

    gettimeofday(&tv, NULL);
    t = tv.tv_sec;
    tm = localtime(&t);
    strftime(ts, sizeof(ts) - 6, "%F_%H:%M:%S", tm);          // Leave 6 bytes for milliseconds and null terminator
    sprintf(ts + strlen(ts), ".%03ld", tv.tv_usec / 1000); // Append milliseconds using sprintf

    if (env.logfile)
    {
        fprintf(filep, "[%s] cpuid:%-5d pid:%-8d comm:%-16s  ",ts, e->cpu, e->pid, e->task);   	
     	if(e->event_type == SOFTIRQ_TYPE)
            fprintf(filep, "soft_irq runtime:   %-8llums si_name:%s\n", e->irq_runtime/1000000, si_vec_names[e->irq_num]);
        else if(e->event_type == SPECTIAL_VECTOR_TYPE)
            fprintf(filep, "special_irq runtime:%-8llums ve_name:%s\n", e->irq_runtime/1000000, special_vec_names[e->irq_num]);    
        else
#if BUILD_WITH_BTF == 1
            fprintf(filep, "hard_irq runtime:   %-8llums hi_name:%s\n", e->irq_runtime/1000000, e->hi_name);
#else
            fprintf(filep, "hard_irq runtime:   %-8llums hi_num:%u\n", e->irq_runtime/1000000, e->irq_num);
#endif
    }
    else
    {
        printf("[%s] cpuid:%-5d pid:%-8d comm:%-16s ", ts, e->cpu, e->pid, e->task);	
        if(e->event_type == SOFTIRQ_TYPE)
            printf("soft_irq runtime:   %-8llums si_name:%s\n", e->irq_runtime/1000000, si_vec_names[e->irq_num]);
        else if(e->event_type == SPECTIAL_VECTOR_TYPE)
            printf("special_irq runtime:%-8llums ve_name:%s\n", e->irq_runtime/1000000, special_vec_names[e->irq_num]);		            		
	else	
#if BUILD_WITH_BTF == 1
            printf("hard_irq runtime:   %-8llums hi_name:%s\n", e->irq_runtime/1000000, e->hi_name);
#else
            printf("hard_irq runtime:   %-8llums hi_num:%u\n", e->irq_runtime/1000000, e->irq_num);
#endif
    }

    if(env.stack){
        print_kernel_stack(filep, env.perf_max_stack_depth, k_sfd, env.logfile, e->k_stack_id, ksyms);
        print_user_stack(filep, env.perf_max_stack_depth, u_sfd, env.logfile, e->u_stack_id, e->tgid, syms_cache);
    }

    if (env.logfile)
        fflush(filep);
}

void handle_lost_events(void *ctx, int cpu, __u64 lost_cnt)
{
    if (env.logfile)
        fprintf(filep, "Lost %llu events on CPU #%d!\n", lost_cnt, cpu);
    else
        printf("Lost %llu events on CPU #%d!\n", lost_cnt, cpu);
}

int main(int argc, char **argv)
{
    static const struct argp argp = {
        .options = opts,
        .parser = parse_arg,
        .doc = argp_program_doc,
    };
    struct perf_buffer *pb = NULL;
    struct irqslower_bpf *obj = NULL;
    int bpf_arg_map_fd;
    struct bpf_arg bpf_arg = {};
    __u32 zero_key = 0;
    int err;
    setlinebuf(stdout);
    err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
    if (err)
        return err;

    if(env.enable_softirq == false && env.enable_hardirq == false)
    {
        env.enable_softirq = true;
        env.enable_hardirq = true;
    }

    if(env.logfile)
    {
        filep = open_log_file(filename);
        if (!filep)
        {
                fprintf(stderr, "Failed to open log file\n");
                return 0;
        }
    }

    libbpf_set_strict_mode(LIBBPF_STRICT_ALL);
    libbpf_set_print(libbpf_print_fn);

    obj = irqslower_bpf__open();
    if (!obj) {
       fprintf(stderr, "failed to open BPF object\n");
       return 1;
    }

    if(env.enable_softirq == false){
        bpf_program__set_autoload(obj->progs.softirq_entry, false);
        bpf_program__set_autoload(obj->progs.softirq_exit, false);
    }
    if(env.enable_hardirq == false){
        bpf_program__set_autoload(obj->progs.irq_handler_entry, false);
        bpf_program__set_autoload(obj->progs.irq_handler_exit, false);
	    bpf_program__set_autoload(obj->progs.call_function_single_entry, false);
        bpf_program__set_autoload(obj->progs.call_function_single_exit, false);
    }

    bpf_map__set_value_size(obj->maps.user_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
    bpf_map__set_max_entries(obj->maps.user_stackmap, env.stack_storage_size);
    bpf_map__set_value_size(obj->maps.kernel_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
    bpf_map__set_max_entries(obj->maps.kernel_stackmap, env.stack_storage_size);

    err = irqslower_bpf__load(obj);
    if (err) {
        fprintf(stderr, "failed to load BPF object: %d\n", err);
        goto cleanup;
    }

    bpf_arg_map_fd = bpf_map__fd(obj->maps.bpf_arg_map);
    bpf_arg.min_n = env.min_us * 1000;
    bpf_arg.stack = env.stack;
    bpf_map_update_elem(bpf_arg_map_fd, &zero_key, &bpf_arg, BPF_ANY);

    if(env.stack)
    {
        ksyms = ksyms__load();
        if (!ksyms)
        {
                fprintf(stderr, "failed to load kallsyms\n");
                goto cleanup;
        }

        syms_cache = syms_cache__new(0);
        if (!syms_cache)
        {
                fprintf(stderr, "failed to create syms_cache\n");
                goto cleanup;
        }

        u_sfd = bpf_map__fd(obj->maps.user_stackmap);
        k_sfd = bpf_map__fd(obj->maps.kernel_stackmap);
    }

    err = irqslower_bpf__attach(obj);
    if (err) {
        fprintf(stderr, "failed to attach BPF programs\n");
	goto cleanup;
    }

    if (env.logfile){
        fprintf(filep, "Tracing irq higher than %llu us\n", env.min_us);
        fflush(filep);
    } else {
        printf("Tracing irq runtime higher than %llu us\n", env.min_us);
    }

    pb = perf_buffer__new(bpf_map__fd(obj->maps.events), 64,
                              handle_event, handle_lost_events, NULL, NULL);
    if (!pb)
    {
        err = -errno;
        fprintf(stderr, "failed to open perf buffer: %d\n", err);
        goto cleanup;
    }

    if (signal(SIGINT, sig_int) == SIG_ERR)
    {
        fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
        err = 1;
        goto cleanup;
    }

    if (signal(SIGTERM, sig_int) == SIG_ERR)
    {
        fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
        err = 1;
        goto cleanup;
    }

    while (!exiting)
    {
        err = perf_buffer__poll(pb, -1);
        if (err < 0 && err != -EINTR)
        {
                fprintf(stderr, "error polling perf buffer: %s\n", strerror(-err));
                goto cleanup;
        }
        /* reset err to return 0 if exiting */
        err = 0;
    }

cleanup:
    perf_buffer__free(pb);
    irqslower_bpf__destroy(obj);
    if (ksyms)
        ksyms__free(ksyms);
    if (syms_cache)
        syms_cache__free(syms_cache);
    if (env.logfile && filep)
    {
        fflush(filep);
        fclose(filep);
    }

	return err != 0;
}
