/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (c) 2024 chudihuang/hilyzhang */
#define BUILD_WITH_BTF 1
#include <argp.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <time.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <err.h>
#include <sys/time.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>
#include <stdbool.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/utsname.h>
#include <linux/version.h>
#include "trace_helpers.h"
#include "noschedule.skel.h"
#include "noschedule.h"

static volatile sig_atomic_t exiting = 0;
FILE *filep = NULL;
char filename[256] = {0};
struct env
{
        __u64 min_us;
        __u64 count;
        __u64 increase_us;
        int pid;
        int target_cg_pid;
        char comm[TASK_COMM_LEN];
        int comm_len;
        int stack_storage_size;
        int perf_max_stack_depth;
        bool runtime;
        bool verbose;
        bool logfile;
        bool set_increase;
} env = {
    .min_us = 10000,
    .count = 0,
    .increase_us = 100000,
    .pid = 0,
    .target_cg_pid = 0,
    .runtime = false,

    .stack_storage_size = 2048,
    .perf_max_stack_depth = 127,
    .logfile = false,
//    .set_increase = true,
};
int u_sfd, k_sfd;
struct ksyms *ksyms = NULL;
struct syms_cache *syms_cache = NULL;
const char *argp_program_version = "noschedule 0.1";
const char *argp_program_bug_address =
    "https://github.com/iovisor/bcc/tree/master/libbpf-tools";
const char argp_program_doc[] =
    "Trace process holding cpu time higher than threshold.\n"
    "\n"
#if BUILD_WITH_BTF == 1
    "USAGE: noschedule [min_us] [-p xxx] [-c xxx] [-k] [-i xxx] [-g xxx] [-f xxx.log] \n"
#else
    "USAGE: noschedule [min_us] [-p xxx] [-c xxx] [-k] [-i xxx] [-f xxx.log] \n"
#endif
    "\n"
    "EXAMPLES:\n"
    "    noschedule               # trace run in sys time higher than 10ms default\n"
    "    noschedule 10000         # trace run in sys time higher than 10000us\n"
    "    noschedule 10000 -r      # trace process runtime higher than 10000 us \n"
    "    noschedule -p xxx        # trace specified process in sys stat\n"
#if BUILD_WITH_BTF == 1
    "    noschedule -g xxx        # trace under same cgroup as the pid xxx\n"
#endif
    "    noschedule -i 200000     # output once every 200000us\n"
    "    noschedule 1000 -c test  # trace the process who's name is test\n"
    "    noschedule -f xxx.log    # log to xxx.log, default output to terminal\n";

static const struct argp_option opts[] = {
    //  -f 󣬳˱Ϣдն⣬Ϣд logfile 
    {"logfile", 'f', "LOGFILE", 0, "logfile for result"},
    {"runtime", 'r', NULL, 0, "trace runtime(user+sys+irq)"},
    { "pid", 'p', "PID", 0, "process PID to trace(include threads)" },
#if BUILD_WITH_BTF == 1
    { "target-cg", 'g', "PID", 0, "trace processes under same cgroup as the pid xxx" },
#endif
    {"increase", 'i', "INCREASE", 0, "set the output once every xxx us "},

    {"comm", 'c', "COMM", 0, "name of the application"},
    {"verbose", 'v', NULL, 0, "verbose debug output"},
    {NULL, 'h', NULL, OPTION_HIDDEN, "show the full help"},
    {},
};

static __u32 get_ubuntu_kernel_version_value(void)
{
        const char *ubuntu_kver_file = "/proc/version_signature";
        __u32 major, minor, patch;
        int ret;
        FILE *f;

        if (faccessat(AT_FDCWD, ubuntu_kver_file, R_OK, AT_EACCESS) != 0)
                return 0;

        f = fopen(ubuntu_kver_file, "re");
        if (!f)
                return 0;

        ret = fscanf(f, "%*s %*s %u.%u.%u\n", &major, &minor, &patch);
        fclose(f);
        if (ret != 3)
                return 0;

        return KERNEL_VERSION(major, minor, patch);
}
static __u32 get_debian_kernel_version_value(struct utsname *info)
{
        __u32 major, minor, patch;
        char *p;

        p = strstr(info->version, "Debian ");
        if (!p) {
                /* This is not a Debian kernel. */
                return 0;
        }

        if (sscanf(p, "Debian %u.%u.%u", &major, &minor, &patch) != 3)
                return 0;

        return KERNEL_VERSION(major, minor, patch);
}
__u32 get_kernel_version_value(void)
{
        __u32 major, minor, patch, version;
        struct utsname info;

        /* Check if this is an Ubuntu kernel. */
        version = get_ubuntu_kernel_version_value();
        if (version != 0)
                return version;

        uname(&info);

        /* Check if this is a Debian kernel. */
        version = get_debian_kernel_version_value(&info);
        if (version != 0)
                return version;

        if (sscanf(info.release, "%u.%u.%u", &major, &minor, &patch) != 3)
                return 0;

        return KERNEL_VERSION(major, minor, patch);
}

bool check_cgroup_namespace(){
        __u32 kernel_version = get_kernel_version_value();
        __u32 compare_kernel_version = KERNEL_VERSION(4,6,0);

        if(kernel_version >= compare_kernel_version){
                return true;
        }

        return false;
}

unsigned int get_namespace_value(int pid, const char *ns_name) {
    char path[PATH_MAX];
    char link_target[PATH_MAX];
    ssize_t len;
    char *start;
    char *end;
    char ns_value_str[32];
    unsigned int ns_value;

    snprintf(path, sizeof(path), "/proc/%d/ns/%s", pid, ns_name);

    len = readlink(path, link_target, sizeof(link_target) - 1);
    if (len == -1) {
        perror("readlink");
        return 0;
    }
    link_target[len] = '\0';

    start = strchr(link_target, '[');
    end = strchr(link_target, ']');
    if (!(start && end && end > start)) {
        printf("Failed to parse namespace value from: %s\n", link_target);
        return 0;
    }
    strncpy(ns_value_str, start + 1, end - start - 1);
    ns_value_str[end - start - 1] = '\0';
    ns_value = strtoul(ns_value_str, NULL, 10);

    return ns_value;
}
static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
        static int pos_args, len, pid, target_cg_pid;
        long long min_us, increase_us;
        bool enable_cgroup_ns;

        switch (key)
        {
        case 'h':
                argp_state_help(state, stderr, ARGP_HELP_STD_HELP);
                break;
        case 'v':
                env.verbose = true;
                break;
        case 'f':
                if (strlen(arg) > 255)
                {
                        fprintf(stderr, "filename too long\n");
                        argp_usage(state);
                }
                strncpy(filename, arg, sizeof(filename)-1);
                env.logfile = true;
                break;
        case 'p':
                errno = 0;
                pid = strtol(arg, NULL, 10);
                if (errno || pid <= 0) {
                        warn("Invalid PID: %s\n", arg);
                        argp_usage(state);
                }
                env.pid = pid;
                break;
#if BUILD_WITH_BTF == 1
        case 'g':
                enable_cgroup_ns = check_cgroup_namespace();
                if(!enable_cgroup_ns){
                        fprintf(stderr, "/proc/<pid>/ns/cgroup likely does not exist, the -g parameter cannot be used\n");
                        exit(0);
                }
                errno = 0;
                target_cg_pid = strtol(arg, NULL, 10);
                if (errno || target_cg_pid <= 0) {
                        warn("Invalid target_cg_pid: %s\n", arg);
                        
                        argp_usage(state);
                }
                env.target_cg_pid = target_cg_pid;
                break;
#endif
        case 'r':
                env.runtime = true;
                break;
        case 'i':
                errno = 0;
                increase_us = strtol(arg, NULL, 10);
                if (errno || increase_us < 0) {
                        warn("Invalid increase_us val: %s\n", arg);
                        
                        argp_usage(state);
                }
                env.increase_us = increase_us;
                //env.set_increase = true;
                break;
        case 'c':
                // env.comm = arg;
                len = strlen(arg) + 1;
                env.comm_len = len > TASK_COMM_LEN ? TASK_COMM_LEN : len;
                strncpy(env.comm, arg, env.comm_len);
                break;
        case ARGP_KEY_ARG:
                if (pos_args++)
                {
                        fprintf(stderr, "Unrecognized positional argument: %s\n", arg);
                        argp_usage(state);
                }
                errno = 0;
                min_us = strtoll(arg, NULL, 10);
                if (errno || min_us <= 0)
                {
                        fprintf(stderr, "Invalid delay (in us): %s\n", arg);
                        argp_usage(state);
                }
                env.min_us = min_us;
                break;
        default:
                return ARGP_ERR_UNKNOWN;
        }
        return 0;
}

FILE *open_log_file(char *filename)
{
        FILE *filep;

        filep = fopen(filename, "w+");
        if (!filep)
        {
                fprintf(stderr, "%s :fopen %s\n", strerror(errno), filename);
                return NULL;
        }

        return filep;
}
static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
{
        if (level == LIBBPF_DEBUG && !env.verbose)
                return 0;
        return vfprintf(stderr, format, args);
}

static void sig_int(int signo)
{
        exiting = 1;
}

void print_kernel_stack(FILE *filep, int perf_max_stack_depth, int sfd, bool logfile, int k_stack_id, struct ksyms *ksyms)
{
        const struct ksym *ksym;
        int i;
        unsigned long *ip;

        ip = calloc(perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        fprintf(stderr, "failed to alloc ip\n");
                return;
        }

        if (bpf_map_lookup_elem(sfd, &k_stack_id, ip) != 0)
        {
                if (logfile)
                        fprintf(filep, "    [Missed Kernel Stack:%d]\n", k_stack_id);
                else
                        fprintf(stderr, "    [Missed Kernel Stack:%d]\n", k_stack_id);
                goto skip_kernel_stack;
        }
        for (i = 0; i < perf_max_stack_depth && ip[i]; i++)
        {
                ksym = ksyms__map_addr(ksyms, ip[i]);
                if (ksym)
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                        else
                                printf("    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                }
                else
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx [unknown]\n", ip[i]);
                        else
                                printf("    0x%lx [unknown]\n", ip[i]);
                }
        }

skip_kernel_stack:
        free(ip);
}

void print_user_stack(FILE *filep, int perf_max_stack_depth, int sfd, bool logfile, int u_stack_id, unsigned long tgid, struct syms_cache *syms_cache)
{
        const struct syms *syms;
        const struct sym *sym;
        int i;
        unsigned long *ip;

        ip = calloc(perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        fprintf(stderr, "failed to alloc ip\n");
                return;
        }

        if (bpf_map_lookup_elem(sfd, &u_stack_id, ip) != 0)
        {
                if (logfile)
                        fprintf(filep, "    [Missed User Stack:%d]\n", u_stack_id);
                else
                        fprintf(stderr, "    [Missed User Stack:%d]\n", u_stack_id);
                goto skip_user_stack;
        }

        syms = syms_cache__get_syms(syms_cache, tgid);
        if (!syms)
        {
                if (logfile)
                        fprintf(filep, "    failed to get syms\n");
                else
                        fprintf(stderr, "    failed to get syms\n");
                goto skip_user_stack;
        }
	for (i = 0; i < perf_max_stack_depth && ip[i]; i++)
        {
                sym = syms__map_addr(syms, ip[i]);
                if (sym)
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx %s\n", ip[i], sym->name);
                        else
                                printf("    0x%lx %s\n", ip[i], sym->name);
                }
                else
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx [unknown]\n", ip[i]);
                        else
                                printf("    0x%lx [unknown]\n", ip[i]);
                }
        }

skip_user_stack:
        free(ip);
}

void handle_event(void *ctx, int cpu, void *data, __u32 data_sz)
{
        const struct event *e = data;
        struct tm *tm;
        char ts[64]; // Increase the size of the ts array
        struct timeval tv;
        time_t t;

        gettimeofday(&tv, NULL);
        t = tv.tv_sec;
        tm = localtime(&t);
        strftime(ts, sizeof(ts) - 6, "%F_%H:%M:%S", tm);          // Leave 6 bytes for milliseconds and null terminator
        sprintf(ts + strlen(ts), ".%03ld", tv.tv_usec / 1000); // Append milliseconds using sprintf

        if (env.logfile)
        {
                fprintf(filep, "[%s]: \n", ts);
                fprintf(filep, "[runcount:%lu]  cpuid:%d  pid:%d  comm: %-16s  stime:%llums  utime:%llums  exec_time:%llums\n", 
                        e->pcount, e->cpu, e->pid, e->task, e->this_stime / 1000000, e->this_utime / 1000000,  e->exec_time / 1000000);
        }
        else
        {
                printf("[%s]: \n", ts);
                printf("[runcount:%lu]  cpuid:%d  pid:%d  comm: %-16s  stime:%llums  utime:%llums  exec_time:%llums\n", 
                        e->pcount, e->cpu, e->pid, e->task, e->this_stime / 1000000, e->this_utime / 1000000, e->exec_time / 1000000);
        }

        print_kernel_stack(filep, env.perf_max_stack_depth, k_sfd, env.logfile, e->k_stack_id, ksyms);
        print_user_stack(filep, env.perf_max_stack_depth, u_sfd, env.logfile, e->u_stack_id, e->tgid, syms_cache);

}

void handle_lost_events(void *ctx, int cpu, __u64 lost_cnt)
{
        if (env.logfile)
                fprintf(filep, "Lost %llu events on CPU #%d!\n", lost_cnt, cpu);
        else
                printf("Lost %llu events on CPU #%d!\n", lost_cnt, cpu);
}

int main(int argc, char **argv)
{
        static const struct argp argp = {
            .options = opts,
            .parser = parse_arg,
            .doc = argp_program_doc,
        };
        struct perf_buffer *pb = NULL;
        struct noschedule_bpf *obj = NULL;
        int bpf_arg_map_fd;
        struct bpf_arg bpf_arg = {};
        int zero_key = 0;
        int err;
        bool enable_kprobe;

        setlinebuf(stdout);
        memset(env.comm, '\0', TASK_COMM_LEN);
        err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
        if (err)
                return err;

        if(env.logfile)
        {
                filep = open_log_file(filename);
                if (!filep)
                {
                        fprintf(stderr, "Failed to open log file\n");
                        return 0;
                }
                setlinebuf(filep);
        }

        libbpf_set_strict_mode(LIBBPF_STRICT_ALL);
        libbpf_set_print(libbpf_print_fn);

        obj = noschedule_bpf__open();
        if (!obj)
        {
                fprintf(stderr, "failed to open BPF object\n");
                return 1;
        }

        enable_kprobe = kprobe_exists("finish_task_switch");
        if(enable_kprobe){
                bpf_program__set_autoload(obj->progs.raw_sched_switch, false);
        } else {
                bpf_program__set_autoload(obj->progs.sched_switch, false);
                bpf_program__set_autoload(obj->progs.trace_finish_task_switch, false);
        }

        bpf_map__set_value_size(obj->maps.user_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
        bpf_map__set_max_entries(obj->maps.user_stackmap, env.stack_storage_size);
        bpf_map__set_value_size(obj->maps.kernel_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
        bpf_map__set_max_entries(obj->maps.kernel_stackmap, env.stack_storage_size);

        err = noschedule_bpf__load(obj);
        if (err)
        {
                fprintf(stderr, "failed to load BPF object: %d\n", err);
                goto cleanup;
        }

        bpf_arg_map_fd = bpf_map__fd(obj->maps.bpf_arg_map);
        bpf_arg.compare_runtime = env.runtime;
       
        bpf_arg.min_ns = env.min_us * 1000;
        bpf_arg.kernel_version = get_kernel_version_value();
        bpf_arg.tgid = env.pid;
        if (env.comm_len && env.comm)
        {
                strncpy((char *)bpf_arg.targ_comm, env.comm, env.comm_len);
                bpf_arg.comm_len = env.comm_len;
        }
        if(bpf_arg.kernel_version == 0)
        {
                fprintf(stderr, "failed to get kernel version\n");
                goto cleanup;
        }
        bpf_arg.increase = env.increase_us * 1000;
        if(env.target_cg_pid)
        {
                bpf_arg.target_cg = get_namespace_value(env.target_cg_pid, "cgroup");
                if(!bpf_arg.target_cg)
                        return 0;
        }
        bpf_map_update_elem(bpf_arg_map_fd, &zero_key, &bpf_arg, BPF_ANY);

        ksyms = ksyms__load();
        if (!ksyms)
        {
                fprintf(stderr, "failed to load kallsyms\n");
                goto cleanup;
        }

        syms_cache = syms_cache__new(0);
        if (!syms_cache)
        {
                fprintf(stderr, "failed to create syms_cache\n");
                goto cleanup;
        }

        u_sfd = bpf_map__fd(obj->maps.user_stackmap);
        k_sfd = bpf_map__fd(obj->maps.kernel_stackmap);

        err = noschedule_bpf__attach(obj);
        if (err)
        {
                fprintf(stderr, "failed to attach BPF programs\n");
                goto cleanup;
        }

        if (env.logfile)
        {
                fprintf(filep, "Tracing process hold cpu higher than %llu us\n", env.min_us);
                // fflush(filep);
        }
        else
                printf("Tracing process hold cpu higher than %llu us\n", env.min_us);

        pb = perf_buffer__new(bpf_map__fd(obj->maps.events), 64,
                              handle_event, handle_lost_events, NULL, NULL);
        if (!pb)
        {
                err = -errno;
                fprintf(stderr, "failed to open perf buffer: %d\n", err);
                goto cleanup;
        }

        if (signal(SIGINT, sig_int) == SIG_ERR)
        {
                fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
                err = 1;
                goto cleanup;
        }

        if (signal(SIGTERM, sig_int) == SIG_ERR)
        {
                fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
                err = 1;
                goto cleanup;
        }

        while (!exiting)
        {
                err = perf_buffer__poll(pb, -1);
                if (err < 0 && err != -EINTR)
                {
                        fprintf(stderr, "error polling perf buffer: %s\n", strerror(-err));
                        goto cleanup;
                }
                /* reset err to return 0 if exiting */
                err = 0;
        }

cleanup:
        perf_buffer__free(pb);
        noschedule_bpf__destroy(obj);
        if (ksyms)
                ksyms__free(ksyms);
        if (syms_cache)
                syms_cache__free(syms_cache);
        if (env.logfile && filep)
        {
                // fflush(filep);
                fclose(filep);
        }

        return err != 0;
}
