/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (c) 2024 chudihuang */
#include <argp.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <time.h>
#include <sys/stat.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <time.h>
#include <err.h>
#include <sys/time.h>
#include <sys/param.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>
#include "syscallslower.skel.h"
#include "trace_helpers.h"
#include "syscall_helpers.h"
#include "syscallslower.h"
static volatile sig_atomic_t exiting = 0;
FILE *filep = NULL;
char filename[256] = {0};
struct env
{
        __u64 min_us;
        __u32 pid;
        int target_cg_pid;
        __u32 tid;
        char comm[TASK_COMM_LEN];
        int comm_len;
        __u32 sysnr[MAX_MATCH_NR];
        __u32 excludesysnr[MAX_MATCH_NR];
        int stack_storage_size;
        int perf_max_stack_depth;
        bool stacktrace;
        bool verbose;
        bool list_syscalls;
        bool filter_sysnr;
        char *cgroupspath;
        bool filter_cg;
        bool logfile;
} env = {
    .pid = 0,
    .tid = 0,
    .target_cg_pid = 0,
    .min_us = 1000,
    .stacktrace = false,
    .stack_storage_size = 2048,
    .filter_sysnr = false,
    .filter_cg = false,
    .logfile = false,
    .perf_max_stack_depth = PERF_MAX_STACK_DEPTH,
};
int ifd, u_sfd, k_sfd;
unsigned int zero_key = 0;
int stackmp_fd;
unsigned long long min_ns;
struct ksyms *ksyms = NULL;
struct syms_cache *syms_cache = NULL;
const char *argp_program_version = "syscallslower 0.1";
const char *argp_program_bug_address =
    "https://github.com/iovisor/bcc/tree/master/libbpf-tools";
const char argp_program_doc[] =
    "Trace syscalls latency.\n"
    "\n"
    "USAGE: syscallslower [--help]\n"
    "\n"
    "EXAMPLES:\n"
    "    syscallslower              # trace process block longer than default 1000 us\n"
    "    syscallslower 5000         # detect with threshold 5ms (default 1ms)\n"
    "    syscallslower -g xxx       # trace the processes in the same cgroup as the xxx\n"
    "    syscallslower -t 185       # trace tid 185 (use for threads only)\n"
    "    syscallslower -p 185       # trace pid 185 process group(include threads)\n"
    "    syscallslower -s           # show stack\n"
    "    syscallslower -l           # Print list of recognized syscalls\n"
    "    syscallslower -e 35,230    # trace syscall 35 and 230 from tracing.\n"
    "    syscallslower -E 35,230    # Exclude syscall 35 and 230 from tracing.\n"
    "    syscallslower -c test      # trace process who's name is test\n"
//    "    syscallslower -C CG        # Trace process under cgroupsPath CG,only support cgroup v2\n"
    "    syscallslower -f xxx.log   # write log to xxx.log (default output to terminal)\n";

static const struct argp_option opts[] = {
    {"pid", 'p', "pid", 0, "Process PID to trace(include threads)"},
    {"tid", 't', "tid", 0, "Thread TID to trace"},
    {"target-cg", 'g', "TARGET-CG", 0, "trace the processes in the same cgroup as the xxx process" },
    {"sysnr", 'e', "sysnr", 0, "Comma-separated list of syscall numbers to trace"},
    {"excludesysnr", 'E', "excludesysnr", 0, "Comma-separated list of exclude syscall numbers to trace"},
    {"comm", 'c', "COMM", 0, "Name of the application"},
    {"stacktrace", 's', NULL, 0, "Show stack trace"},
    {"list", 'l', NULL, 0, "Print list of recognized syscalls and exit"},
    {"verbose", 'v', NULL, 0, "Verbose debug output"},
    {"logfile", 'f', "LOGFILE", 0, "logfile for result"},
    {NULL, 'h', NULL, OPTION_HIDDEN, "Show the full help"},
    {},
};

FILE *open_log_file(char *filename)
{
        FILE *filep;

        filep = fopen(filename, "w+");
        if (!filep)
        {
                fprintf(stderr, "%s :fopen %s\n", strerror(errno), filename);
                return NULL;
        }

        return filep;
}

static int get_sysnrs(const char *arg, __u32 max_size, __u32 *ret, int min, int max)
{
        const char *argp = arg;
        int sz = 0;
        char *end;
        long val;
        while (sz < max_size)
        {
                errno = 0;
                val = strtol(argp, &end, 10);
                if (errno)
                {
                        return -1;
                }
                else if (end == arg || val < min || val > max)
                {
                        return -1;
                }
                ret[sz++] = val;
                if (*end == 0)
                        break;
                argp = end + 1;
        }

        return 0;
}

static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
        static int pos_args, len, target_cg_pid;
        __u32 sysnr_size;
        int err;
        long long min_us;

        switch (key)
        {
        case 'h':
                argp_state_help(state, stderr, ARGP_HELP_STD_HELP);
                break;
        case 'v':
                env.verbose = true;
                break;
        case 's':
                env.stacktrace = true;
                break;
        case 'p':
                errno = 0;
                env.pid = strtol(arg, NULL, 10);
                if (errno)
                {
                        fprintf(stderr, "invalid PID: %s\n", arg);
                        argp_usage(state);
                }
                break;
        case 'g':
                errno = 0;
                target_cg_pid = strtol(arg, NULL, 10);
                if (errno || target_cg_pid <= 0) {
                        warn("Invalid target_cg_pid: %s\n", arg);
                        argp_usage(state);
                }
                env.target_cg_pid = target_cg_pid;
                break;
        case 't':
                errno = 0;
                env.tid = strtol(arg, NULL, 10);
                if (errno)
                {
                        fprintf(stderr, "invalid PID: %s\n", arg);
                        argp_usage(state);
                }
                break;
        case 'f':
                if (strlen(arg) > 255)
                {
                        fprintf(stderr, "filename too long\n");
                        argp_usage(state);
                }
                strncpy(filename, arg, sizeof(filename)-1);
                env.logfile = true;
                break;
        case 'l':
                env.list_syscalls = true;
                break;
        case 'e':
                sysnr_size = MIN(MAX_NR, MAX_MATCH_NR);
                err = get_sysnrs(arg, sysnr_size, env.sysnr, 0, USHRT_MAX);
                if (err)
                {
                        fprintf(stderr, "Invalid syscall nr: %s\n", arg);
                        return -1;
                }
                env.filter_sysnr = true;
                break;
        case 'E':
                sysnr_size = MIN(MAX_NR, MAX_MATCH_NR);
                err = get_sysnrs(arg, sysnr_size, env.excludesysnr, 0, USHRT_MAX);
                if (err)
                {
                        fprintf(stderr, "Invalid syscall nr: %s\n", arg);
                        return -1;
                }
                break;
        case 'c':
                len = strlen(arg) + 1;
                env.comm_len = len > TASK_COMM_LEN ? TASK_COMM_LEN : len;
                strncpy(env.comm, arg, env.comm_len);
                break;
        case 'C':
                env.cgroupspath = arg;
                env.filter_cg = true;
                break;
        case ARGP_KEY_ARG:
                if (pos_args++)
                {
                        fprintf(stderr,
                                "Unrecognized positional argument: %s\n", arg);
                        argp_usage(state);
                }
                errno = 0;
                min_us = strtoll(arg, NULL, 10);
                if (errno || min_us <= 0)
                {
                        fprintf(stderr, "Invalid delay (in us): %s\n", arg);
                        argp_usage(state);
                }
                env.min_us = min_us;
                break;
        default:
                return ARGP_ERR_UNKNOWN;
        }
        return 0;
}

static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
{
        if (level == LIBBPF_DEBUG && !env.verbose)
                return 0;
        return vfprintf(stderr, format, args);
}

static void sig_int(int signo)
{
        exiting = 1;
}

void print_kernel_stack(FILE *filep, int perf_max_stack_depth, int sfd, bool logfile, int k_stack_id, struct ksyms *ksyms)
{
        const struct ksym *ksym;
        int i;
        unsigned long *ip;

        ip = calloc(perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        fprintf(stderr, "failed to alloc ip\n");
                return;
        }

        if (bpf_map_lookup_elem(sfd, &k_stack_id, ip) != 0)
        {
                if (logfile)
                        fprintf(filep, "    [Missed Kernel Stack:%d]\n", k_stack_id);
                else
                        fprintf(stderr, "    [Missed Kernel Stack:%d]\n", k_stack_id);
                goto skip_kernel_stack;
        }

        for (i = 0; i < perf_max_stack_depth && ip[i]; i++)
        {
                ksym = ksyms__map_addr(ksyms, ip[i]);
                if (ksym)
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                        else
                                printf("    0x%lx %s+0x%lx\n", ip[i], ksym->name, ip[i] - ksym->addr);
                }
                else
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx [unknown]\n", ip[i]);
                        else
                                printf("    0x%lx [unknown]\n", ip[i]);
	        }
        }

skip_kernel_stack:
        free(ip);
}	

void print_user_stack(FILE *filep, int perf_max_stack_depth, int sfd, bool logfile, int u_stack_id, unsigned long pid, struct syms_cache *syms_cache)
{
        const struct syms *syms;
        const struct sym *sym;
        int i;
        unsigned long *ip;

        ip = calloc(perf_max_stack_depth, sizeof(*ip));
        if (!ip)
        {
                if (logfile)
                        fprintf(filep, "failed to alloc ip\n");
                else
                        fprintf(stderr, "failed to alloc ip\n");
                return;
        }

        if (bpf_map_lookup_elem(sfd, &u_stack_id, ip) != 0)
        {
                if (logfile)
                        fprintf(filep, "    [Missed User Stack:%d]\n", u_stack_id);
                else
                        fprintf(stderr, "    [Missed User Stack:%d]\n", u_stack_id);
                goto skip_user_stack;
        }

        syms = syms_cache__get_syms(syms_cache, pid);
        if (!syms)
        {
                if (logfile)
                        fprintf(filep, "    failed to get syms\n");
                else
                        fprintf(stderr, "    failed to get syms\n");
                goto skip_user_stack;
        }
        for (i = 0; i < perf_max_stack_depth && ip[i]; i++)
        {
                sym = syms__map_addr(syms, ip[i]);
                if (sym)
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx %s\n", ip[i], sym->name);
                        else
                                printf("    0x%lx %s\n", ip[i], sym->name);
                }
		 else
                {
                        if (logfile)
                                fprintf(filep, "    0x%lx [unknown]\n", ip[i]);
                        else
                                printf("    0x%lx [unknown]\n", ip[i]);
                }
        }

skip_user_stack:
        free(ip);
}

void handle_event(void *ctx, int cpu, void *data, __u32 data_sz)
{
        const struct event *e = data;
        struct tm *tm;
        char ts[64]; // Increase the size of the ts array
        struct timeval tv;
        time_t t;

        gettimeofday(&tv, NULL);
        t = tv.tv_sec;
        tm = localtime(&t);
        strftime(ts, sizeof(ts) - 6, "%F_%H:%M:%S", tm);       // Leave 6 bytes for milliseconds and null terminator
        sprintf(ts + strlen(ts), ".%03ld", tv.tv_usec / 1000); // Append milliseconds using sprintf

        char syscallbuf[2 * TASK_COMM_LEN];
        syscall_name(e->sysid, syscallbuf, sizeof(syscallbuf));
        if (env.logfile)
                fprintf(filep, "%-24s %-12llu %-10llu %-10llu %-10llu %-10llu %-10llu %-10llu %-10ld %-10s %u(%s)\n",
                        ts, e->delay / (1000), e->realtime / (1000),
                        e->itime / (1000), e->vtime / (1000),
                        e->stime / (1000), e->nvcsw, e->nivcsw, e->state, syscallbuf, e->pid, e->comm);
        else
                printf("%-24s %-12llu %-10llu %-10llu %-10llu %-10llu %-10llu %-10llu %-10ld %-10s %u(%s)\n",
                       ts, e->delay / (1000), e->realtime / (1000),
                       e->itime / (1000), e->vtime / (1000),
                       e->stime / (1000), e->nvcsw, e->nivcsw, e->state, syscallbuf, e->pid, e->comm);

        if (env.stacktrace)
        {
                print_kernel_stack(filep, env.perf_max_stack_depth, k_sfd, env.logfile, e->k_stack_id, ksyms);
                print_user_stack(filep, env.perf_max_stack_depth, u_sfd, env.logfile, e->u_stack_id, e->tgid, syms_cache);
        }
        
        if (env.logfile)
                fflush(filep);
}

unsigned int get_namespace_value(int pid, const char *ns_name) {
    char path[PATH_MAX];
    char link_target[PATH_MAX];
    ssize_t len;
    char *start;
    char *end;
    char ns_value_str[32];
    unsigned int ns_value;

    snprintf(path, sizeof(path), "/proc/%d/ns/%s", pid, ns_name);

    len = readlink(path, link_target, sizeof(link_target) - 1);
    if (len == -1) {
        perror("readlink");
        return 0;
    }
    link_target[len] = '\0';

    start = strchr(link_target, '[');
    end = strchr(link_target, ']');
    if (!(start && end && end > start)) {
        printf("Failed to parse namespace value from: %s\n", link_target);
        return 0;
    }
    strncpy(ns_value_str, start + 1, end - start - 1);
    ns_value_str[end - start - 1] = '\0';
    ns_value = strtoul(ns_value_str, NULL, 10);

    return ns_value;
}
void handle_lost_events(void *ctx, int cpu, __u64 lost_cnt)
{
        printf("Lost %llu events on CPU #%d!\n", lost_cnt, cpu);
}

int main(int argc, char **argv)
{
        static const struct argp argp = {
            .options = opts,
            .parser = parse_arg,
            .doc = argp_program_doc,
        };
        struct arg_val arg_val = {};
        struct perf_buffer *pb = NULL;
        struct syscallslower_bpf *obj;
        int err, i = 0, arg_fd, sysnr_fd, exclude_fd, sysnr_size;
        int idx, cg_map_fd,cgfd = -1;
        setlinebuf(stdout);
        init_syscall_names();
        memset(env.comm, '\0', TASK_COMM_LEN);
        memset(env.sysnr, -1, sizeof(env.sysnr));
        memset(env.excludesysnr, -1, sizeof(env.excludesysnr));
        err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
        if (err)
                goto free_names;
        if (env.list_syscalls)
        {
                list_syscalls();
                goto free_names;
        }

        if(env.logfile)
        {
                filep = open_log_file(filename);
                if (!filep)
                {
                        fprintf(stderr, "Failed to open log file\n");
                        return 0;
                }
        }

        libbpf_set_strict_mode(LIBBPF_STRICT_ALL);
        libbpf_set_print(libbpf_print_fn);

        obj = syscallslower_bpf__open();
        if (!obj)
        {
                fprintf(stderr, "failed to open BPF object\n");
                return 1;
        }
        bpf_map__set_value_size(obj->maps.user_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
        bpf_map__set_max_entries(obj->maps.user_stackmap, env.stack_storage_size);
        bpf_map__set_value_size(obj->maps.kernel_stackmap, env.perf_max_stack_depth * sizeof(unsigned long));
        bpf_map__set_max_entries(obj->maps.kernel_stackmap, env.stack_storage_size);
        err = syscallslower_bpf__load(obj);
        if (err)
        {
                fprintf(stderr, "failed to load BPF object: %d\n", err);
                goto cleanup;
        }

        ksyms = ksyms__load();
        if (!ksyms)
        {
                fprintf(stderr, "failed to load kallsyms\n");
                goto cleanup;
        }

        syms_cache = syms_cache__new(0);
        if (!syms_cache)
        {
                fprintf(stderr, "failed to create syms_cache\n");
                goto cleanup;
        }

        u_sfd = bpf_map__fd(obj->maps.user_stackmap);
        k_sfd = bpf_map__fd(obj->maps.kernel_stackmap);
        arg_fd = bpf_map__fd(obj->maps.arg_map);
        if (env.comm_len && env.comm)
        {
                strncpy((char *)arg_val.targ_comm, env.comm, env.comm_len);
                arg_val.comm_len = env.comm_len;
        }
        arg_val.thresh = env.min_us * 1000;
        arg_val.pid = env.pid;
        arg_val.tid = env.tid;
        arg_val.filter_sysnr = env.filter_sysnr;
        arg_val.filter_cg = env.filter_cg;
        if(env.target_cg_pid)
        {
                arg_val.target_cg = get_namespace_value(env.target_cg_pid, "cgroup");
                if(!arg_val.target_cg)
                        return 0;
        }
        err = bpf_map_update_elem(arg_fd, &zero_key, &arg_val, 0);
        if (err)
        {
                fprintf(stderr, "Failed to update arg_map\n");
                goto cleanup;
        }
        /* update cgroup v2 path fd to map */
        if (env.filter_cg)
        {
                idx = 0;
                cg_map_fd = bpf_map__fd(obj->maps.cgroup_map);
                cgfd = open(env.cgroupspath, O_RDONLY);
                if (cgfd < 0)
                {
                        fprintf(stderr, "Failed opening Cgroup path: %s\n", env.cgroupspath);
                        goto cleanup;
                }
                if (bpf_map_update_elem(cg_map_fd, &idx, &cgfd, BPF_ANY))
                {
                        fprintf(stderr, "Failed adding target cgroup(%s) to map\n", env.cgroupspath);
                        goto cleanup;
                }
        }
        sysnr_fd = bpf_map__fd(obj->maps.sysnr_map);
        sysnr_size = MIN(MAX_NR, MAX_MATCH_NR);
        while (i < sysnr_size)
        {
                err = bpf_map_update_elem(sysnr_fd, &env.sysnr[i], &env.sysnr[i], 0);
                if (err)
                {
                        fprintf(stderr, "Failed to update sysnr map\n");
                        goto cleanup;
                }
                i++;
        }
        i = 0;
        exclude_fd = bpf_map__fd(obj->maps.exclude_map);
        sysnr_size = MIN(MAX_NR, MAX_MATCH_NR);
        while (i < sysnr_size)
        {
                err = bpf_map_update_elem(exclude_fd, &env.excludesysnr[i], &env.excludesysnr[i], 0);
                if (err)
                {
                        fprintf(stderr, "Failed to update excludesysnr map %d\n", i);
                        goto cleanup;
                }
                i++;
        }

        err = syscallslower_bpf__attach(obj);
        if (err)
        {
                fprintf(filep, "failed to attach BPF programs\n");
                goto cleanup;
        }
        if (env.logfile)
                fprintf(filep, "%-24s %-12s %-12s %-12s %-12s %-12s %-12s %-12s %-12s %-10s\n",
                        "TIME(syscall)", "DELAY(us)", "REAL(us)", "WAIT(us)", "SLEEP(us)",
                        "SYS(us)", "vcsw", "ivcsw", "syscall", "pid(comm)");
        else
                printf("%-24s %-12s %-10s %-10s %-10s %-10s %-10s %-10s %-10s %-10s %-10s\n",
                       "TIME(syscall)", "DELAY(us)", "REAL(us)", "WAIT(us)", "SLEEP(us)",
                       "SYS(us)", "vcsw", "ivcsw", "state", "syscall", "pid(comm)");
        pb = perf_buffer__new(bpf_map__fd(obj->maps.events), 64,
                              handle_event, handle_lost_events, NULL, NULL);
        if (!pb)
        {
                err = -errno;
                fprintf(filep, "failed to open perf buffer: %d\n", err);
                goto cleanup;
        }

        if (signal(SIGINT, sig_int) == SIG_ERR)
        {
                fprintf(filep, "can't set signal handler: %s\n", strerror(errno));
                err = 1;
                goto cleanup;
        }
        if (signal(SIGTERM, sig_int) == SIG_ERR)
        {
                fprintf(stderr, "can't set signal handler: %s\n", strerror(errno));
                err = 1;
                goto cleanup;
        }
        fflush(filep);
        while (!exiting)
        {
                err = perf_buffer__poll(pb, -1);
                if (err < 0 && err != -EINTR)
                {
                        fprintf(filep, "error polling perf buffer: %s\n", strerror(-err));
                        goto cleanup;
                }
                /* reset err to return 0 if exiting */
                err = 0;
        }

cleanup:
        perf_buffer__free(pb);
        syscallslower_bpf__destroy(obj);
        if (ksyms)
                ksyms__free(ksyms);
        if (syms_cache)
                syms_cache__free(syms_cache);
        if (filep)
        {
                fflush(filep);
                fclose(filep);
        }
free_names:
        free_syscall_names();
        return err != 0;
}
